"""Test for Visual Protocol."""

import logging
import os
import pickle as pkl
from typing import Union, Tuple, Dict, List
import ubelt as ub

from sail_on_client.harness.test_and_evaluation_harness import (
    TestAndEvaluationHarnessType,
)
from sail_on_client.protocol.visual_round import VisualRound
from sail_on_client.utils.decorators import skip_stage
from sail_on_client.protocol.ond_dataclasses import (
    AlgorithmAttributes as ONDAlgorithmAttributes,
)
from sail_on_client.protocol.condda_dataclasses import (
    AlgorithmAttributes as CONDDAAlgorithmAttributes,
)


log = logging.getLogger(__name__)


class VisualTest:
    """Class representing test in visual protocol."""

    def __init__(
        self,
        algorithm_attributes: Union[ONDAlgorithmAttributes, CONDDAAlgorithmAttributes],
        data_root: str,
        domain: str,
        feature_dir: str,
        harness: TestAndEvaluationHarnessType,
        save_dir: str,
        session_id: str,
        skip_stages: List[str],
        use_consolidated_features: bool,
        use_saved_features: bool,
    ) -> None:
        """
        Construct visual test.

        Args:
            algorithm_attributes: An instance of algorithm_attributes
            data_root: Root directory for the dataset
            domain: Name of the domain for the test
            feature_dir: Directory to save features
            harness: An Instance of harness used for T&E
            save_dir: The directory where features are saved
            session_id: Session identifier for the test
            skip_stages: List of stages that would be skipped
            use_consolidated_features: Flag for using consolidated features
            use_saved_features: Flag for using saved features

        Returns:
            None
        """
        self.algorithm_attributes = algorithm_attributes
        self.data_root = data_root
        self.domain = domain
        self.harness = harness
        self.feature_dir = feature_dir
        self.save_dir = save_dir
        self.session_id = session_id
        self.skip_stages = skip_stages
        self.use_consolidated_features = use_consolidated_features
        self.use_saved_features = use_saved_features

    def _restore_features(self, test_id: str) -> Tuple[Dict, Dict]:
        """
        Private function to restore features.

        Args:
           test_id: An identifier for the test

        Returns:
            Tuple of dictionary with features and logits obtained from the feature extractor
        """
        features_dict: Dict = {}
        logit_dict: Dict = {}
        algorithm_name = self.algorithm_attributes.name
        if self.use_saved_features:
            if os.path.isdir(self.feature_dir):
                if self.use_consolidated_features:
                    feature_fname = f"{algorithm_name}_features.pkl"
                else:
                    feature_fname = f"{test_id}_{algorithm_name}_features.pkl"
                feature_path = os.path.join(self.feature_dir, feature_fname)
                test_features = pkl.load(open(feature_path, "rb"))
            else:
                test_features = pkl.load(open(self.feature_dir, "rb"))
            features_dict = test_features["features_dict"]
            logit_dict = test_features["logit_dict"]
        return features_dict, logit_dict

    def _aggregate_features_across_round(
        self, round_instance: VisualRound, feature_dict: Dict, logit_dict: Dict
    ) -> Tuple[Dict, Dict]:
        """
        Aggregate features across multiple rounds.

        Args:
            round_instance: Instance of ond round
            feature_dict: Aggregated features until this function was called
            logit_dict: Aggregated logit until this function was called

        Return:
            Tuple of features and logits with features and logits from the round
        """
        feature_dict.update(getattr(round_instance, "rfeature_dict", {}))
        logit_dict.update(getattr(round_instance, "rlogit_dict", {}))
        return feature_dict, logit_dict

    @skip_stage("SaveFeatures")
    def _save_features(
        self, test_id: str, feature_dict: Dict, logit_dict: Dict
    ) -> None:
        """
        Save features for a test.

        Args:
            test_id: An identifier for the test
            feature_dict: Features for the test
            logit_dict: Logit for the test

        Return:
            None
        """
        ub.ensuredir(self.feature_dir)
        algorithm_name = self.algorithm_attributes.name
        feature_path = os.path.join(
            self.feature_dir, f"{test_id}_{algorithm_name}_features.pkl"
        )
        log.info(f"Saving features in {feature_path}")
        with open(feature_path, "wb") as f:
            pkl.dump({"features_dict": feature_dict, "logit_dict": logit_dict}, f)



"""Test for OND."""

import logging
from itertools import count
from typing import Union, Dict, Any, List

from sail_on_client.protocol.ond_dataclasses import (
    AlgorithmAttributes,
    InitializeParams,
    NoveltyCharacterizationParams,
)
from sail_on_client.protocol.ond_round import ONDRound
from sail_on_client.protocol.visual_test import VisualTest
from sail_on_client.feedback import create_feedback_instance, feedback_type
from sail_on_client.harness.test_and_evaluation_harness import (
    TestAndEvaluationHarnessType,
)
from sail_on_client.utils.utils import safe_remove
from sail_on_client.utils.decorators import skip_stage
from sail_on_client.errors import RoundError


log = logging.getLogger(__name__)


class ONDTest(VisualTest):
    """Class Representing OND test."""

    def __init__(
        self,
        algorithm_attributes: AlgorithmAttributes,
        data_root: str,
        domain: str,
        feedback_type: str,
        feature_dir: str,
        harness: TestAndEvaluationHarnessType,
        save_dir: str,
        session_id: str,
        skip_stages: List[str],
        use_consolidated_features: bool,
        use_saved_features: bool,
    ) -> None:
        """
        Construct test for OND.

        Args:
            algorithm_attributes: An instance of algorithm_attributes
            data_root: Root directory for the dataset
            domain: Name of the domain for the test
            feedback_type: Type of feedback used in the test
            harness: An Instance of harness used for T&E
            save_dir: The directory where features are saved
            session_id: Session identifier for the test
            skip_stages: List of stages that would be skipped
            use_consolidated_features: Flag for using consolidated features
            use_saved_features: Flag for using saved features

        Returns:
            None
        """
        super().__init__(
            algorithm_attributes,
            data_root,
            domain,
            feature_dir,
            harness,
            save_dir,
            session_id,
            skip_stages,
            use_consolidated_features,
            use_saved_features,
        )
        self.feedback_type = feedback_type

    @skip_stage("CreateFeedbackInstance")
    def _create_feedback_instance(
        self, test_id: str, feedback_max_ids: int
    ) -> feedback_type:
        """
        Private function for creating feedback object.

        Args:
           test_id: An identifier for the test
           feedback_max_ids: Budget provided in metadata

        Return:
            An instance of feedback for the domain
        """
        log.info("Creating Feedback object")
        if feedback_max_ids == 0:
            log.warn(
                """feedback_max_ids was missing from metadata, thus setting
                        the feedback budget to 0"""
            )
        feedback_params = {
            "first_budget": feedback_max_ids,
            "income_per_batch": feedback_max_ids,
            "maximum_budget": feedback_max_ids,
            "interface": self.harness,
            "session_id": self.session_id,
            "test_id": test_id,
            "feedback_type": self.feedback_type,
        }
        feedback_instance = create_feedback_instance(self.domain, feedback_params)
        return feedback_instance

    @skip_stage("NoveltyCharacterization")
    def _run_novelty_characterization(
        self, algorithm: Any, nc_params: NoveltyCharacterizationParams, test_id: str
    ) -> None:
        characterization_results = algorithm.execute(
            nc_params.get_toolset(), "NoveltyCharacterization"
        )
        if characterization_results:
            if isinstance(characterization_results, dict):
                self.harness.post_results(
                    characterization_results, test_id, 0, self.session_id
                )
            else:
                results = {"characterization": characterization_results}
                self.harness.post_results(results, test_id, 0, self.session_id)
        else:
            log.warn("No characterization result provided by the algorithm")

    def __call__(self, test_id: str) -> Union[Dict, None]:
        """
        Core logic for running test in OND.

        Args:
            test_id: An identifier for the test

        Returns:
            Score for the test
        """
        metadata = self.harness.get_test_metadata(self.session_id, test_id)
        redlight_instance = metadata.get("red_light", "")
        feedback_max_ids = metadata.get("feedback_max_ids", 0)
        pre_novelty_batches = metadata.get("pre_novelty_batches", 0)
        # Initialize feedback object for the domains
        feedback_instance = self._create_feedback_instance(test_id, feedback_max_ids)

        # Initialize algorithm
        algorithm_instance = self.algorithm_attributes.instance
        algorithm_parameters = self.algorithm_attributes.parameters
        algorithm_init_params = InitializeParams(
            algorithm_parameters,
            self.session_id,
            test_id,
            pre_novelty_batches,
            feedback_instance,
        )
        algorithm_instance.execute(algorithm_init_params.get_toolset(), "Initialize")

        # Restore features
        features_dict, logit_dict = self._restore_features(test_id)

        # Initialize Round
        round_instance = ONDRound(
            algorithm_instance,
            self.data_root,
            features_dict,
            self.harness,
            logit_dict,
            redlight_instance,
            self.session_id,
            self.skip_stages,
            test_id,
        )
        aggregated_features_dict: Dict = {}
        aggregated_logit_dict: Dict = {}
        test_score = {}
        test_instances = []
        # Run algorithm for multiple rounds
        for round_id in count(0):
            log.info(f"Start round: {round_id}")
            # see if there is another round available
            try:
                dataset = self.harness.dataset_request(
                    test_id, round_id, self.session_id
                )
            except RoundError:
                # no more rounds available, this test is done.
                break
            round_score = round_instance(dataset, round_id)
            test_instances.extend(ONDRound.get_instance_ids(dataset))
            if round_score:
                test_score[f"Round {round_id}"] = round_score
            (
                aggregated_features_dict,
                aggregated_logit_dict,
            ) = self._aggregate_features_across_round(
                round_instance, aggregated_features_dict, aggregated_logit_dict
            )
            # cleanup the dataset file for the round
            safe_remove(dataset)
            log.info(f"Round complete: {round_id}")
        nc_params = NoveltyCharacterizationParams(test_instances)
        self._run_novelty_characterization(algorithm_instance, nc_params, test_id)
        self.harness.complete_test(self.session_id, test_id)
        self._save_features(test_id, aggregated_features_dict, aggregated_logit_dict)
        return test_score



"""Json Encoder for Numpy Array."""

import json
import numpy as np
from typing import Any


class NumpyEncoder(json.JSONEncoder):
    """An encoder to convert numpy data types to python primitives."""

    def default(self, obj: Any) -> Any:
        """Defaults for numpy types."""
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.bool_):
            return bool(obj)
        else:
            return super(NumpyEncoder, self).default(obj)



"""Abstract interface for an agent that works for visual agent."""

from smqtk_core import Configurable, Pluggable

from typing import Dict, Tuple, Any, TypeVar
from abc import abstractmethod

VisualAgentType = TypeVar("VisualAgentType", bound="VisualAgent")


class VisualAgent(Configurable, Pluggable):
    """Abstract class for Visual Agent."""

    @classmethod
    def is_usable(cls) -> bool:
        """Determine if this class with be detected by SMQTK's plugin."""
        return True

    def get_config(self) -> Dict:
        """Return a default configuration dictionary."""
        return {}

    @abstractmethod
    def execute(self, toolset: Dict, step_descriptor: str) -> Any:
        """
        Execute method used by the protocol to run different steps associated with the algorithm.

        Args:
            toolset (dict): Dictionary containing parameters for different steps
            step_descriptor (str): Name of the step
        """
        pass

    @abstractmethod
    def feature_extraction(self, fe_toolset: Dict) -> Tuple[Dict, Dict]:
        """
        Abstract method for feature extraction.

        Args:
            fe_toolset: Parameters for feature extraction

        Returns:
            Tuple of dictionary with features and logits.
        """
        pass

    @abstractmethod
    def world_detection(self, wd_toolset: Dict) -> str:
        """
        Abstract method for detecting that the world has changed.

        Args:
            wd_toolset: Parameters for feature extraction

        Returns:
            Path to results for detecting that the world has changed.
        """
        pass



"""Image Classification Class for metrics for sail-on."""

from sail_on_client.evaluate.program_metrics import ProgramMetrics
from sail_on_client.evaluate.metrics import m_acc, m_num, m_ndp, m_num_stats
from sail_on_client.evaluate.metrics import m_ndp_failed_reaction
from sail_on_client.evaluate.metrics import m_accuracy_on_novel
from sail_on_client.evaluate.utils import topk_accuracy

import numpy as np
from pandas import DataFrame
import pandas as pd

from typing import Dict


class ImageClassificationMetrics(ProgramMetrics):
    """Image Classification program metric class."""

    def __init__(
        self, protocol: str, image_id: int, detection: int, classification: int
    ) -> None:
        """
        Initialize.

        Args:
            protocol: Name of the protocol.
            image_id: Column id for image
            detection: Column id for predicting sample wise world detection
            classification: Column id for predicting sample wise classes

        Returns:
            None
        """
        super().__init__(protocol)
        self.image_id = image_id
        self.detection_id = detection
        self.classification_id = classification
        self.novel_id = -1  # Added only for fixing typechecking problems

    def m_acc(
        self,
        gt_novel: DataFrame,
        p_class: DataFrame,
        gt_class: DataFrame,
        round_size: int,
        asymptotic_start_round: int,
    ) -> Dict:
        """
        m_acc function.

        Args:
            gt_novel: ground truth detections (Dimension: [img X detection])
            p_class: class predictions (Dimension: [img X prob that sample is novel, prob of 88 known classes])
            gt_class: ground truth classes (Dimension: [img X detection, classification])
            round_size: size of the round
            asymptotic_start_round: asymptotic samples considered for computing metrics

        Returns:
            Dictionary containing top1, top3 accuracy over the test, pre and post novelty.
        """
        class_prob = p_class.iloc[:, range(1, p_class.shape[1])].to_numpy()
        gt_class_idx = gt_class.to_numpy()
        return m_acc(
            gt_novel, class_prob, gt_class_idx, round_size, asymptotic_start_round
        )

    def m_acc_round_wise(
        self, p_class: DataFrame, gt_class: DataFrame, round_id: int
    ) -> Dict:
        """
        m_acc_round_wise function.

        Args:
            p_class: class predictions
            gt_class: ground truth classes
            round_id: round identifier

        Returns:
            Dictionary containing top1, top3 accuracy for a round
        """
        class_prob = p_class.iloc[:, range(1, p_class.shape[1])].to_numpy()
        gt_class_idx = gt_class.to_numpy()
        top1_acc = topk_accuracy(class_prob, gt_class_idx, k=1)
        top3_acc = topk_accuracy(class_prob, gt_class_idx, k=3)
        return {
            f"top1_accuracy_round_{round_id}": top1_acc,
            f"top3_accuracy_round_{round_id}": top3_acc,
        }

    def m_num(self, p_novel: DataFrame, gt_novel: DataFrame) -> Dict:
        """
        m_num function.

        A Program Metric where the number of samples needed for detecting novelty.
        The method computes the number of GT novel samples needed to predict the first true positive.

        Args:
            p_novel: detection predictions (Dimension: [img X novel])
                Nx1 vector with each element corresponding to probability of novelty
            gt_novel: ground truth detections (Dimension: [img X detection])
                Nx1 vector with each element 0 (not novel) or 1 (novel)

        Returns:
            Difference between the novelty introduction and predicting change in world.
        """
        return m_num(p_novel, gt_novel)

    def m_num_stats(self, p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
        """
        Program Metric.

        Number of samples needed for detecting novelty.  The method computes number of GT novel
            samples needed to predict the first true positive.

        Args:
            p_novel: detection predictions (Dimension: [img X novel])
                Nx1 vector with each element corresponding to probability of novelty
            gt_novel: ground truth detections (Dimension: [img X detection])
                Nx1 vector with each element 0 (not novel) or 1 (novel)

        Returns:
            Dictionary containing indices for novelty introduction and change in world prediction.
        """
        return m_num_stats(p_novel, gt_novel)

    def m_ndp(
        self, p_novel: np.ndarray, gt_novel: np.ndarray, mode: str = "full_test"
    ) -> Dict:
        """
        Novelty Detection Performance: Program Metric.

            Novelty detection performance.  The method computes per-sample novelty detection performance.

        Args:
            p_novel: detection predictions (Dimension: [img X novel])
                Nx1 vector with each element corresponding to probability of it being novel
            gt_novel: ground truth detections (Dimension: [img X detection])
                Nx1 vector with each element 0 (not novel) or 1 (novel)
            mode: the mode to compute the test.  if 'full_test' computes on all test samples,
                if 'post_novelty' computes from first GT novel sample.  If 'pre_novelty', only calculate
                before first novel sample.

        Returns:
            Dictionary containing novelty detection performance over the test.
        """
        return m_ndp(p_novel, gt_novel, mode="full_test")

    def m_ndp_pre(self, p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
        """
        Novelty Detection Performance Pre Red Light.  m_ndp_pre function.

        See :func:`~sail-on-client.evaluation.ImageClassificationMetrics.m_ndp` with
            post_novelty.  This computes to the first GT novel sample.  It really isn't useful
            and is just added for completion.  Should always be 0 since no possible TP.

        Args:
            p_novel: detection predictions (Dimension: [img X novel])
            gt_novel: ground truth detections (Dimension: [img X detection])

        Returns:
            Dictionary containing detection performance pre novelty.
        """
        return m_ndp(p_novel, gt_novel, mode="pre_novelty")

    def m_ndp_post(self, p_novel: np.ndarray, gt_novel: np.ndarray) -> Dict:
        """
        Novelty Detection Performance Post Red Light.  m_ndp_post function.

        See :func:`~sail-on-client.evaluation.ImageClassificationMetrics.m_ndp` with
            post_novelty.  This computes from the first GT novel sample

        Args:
            p_novel: detection predictions (Dimension: [img X novel])
            gt_novel: ground truth detections (Dimension: [img X detection])

        Returns:
            Dictionary containing detection performance post novelty.
        """
        return m_ndp(p_novel, gt_novel, mode="post_novelty")

    def m_ndp_failed_reaction(
        self,
        p_novel: DataFrame,
        gt_novel: DataFrame,
        p_class: DataFrame,
        gt_class: DataFrame,
        mode: str = "full_test",
    ) -> Dict:
        """
        Additional Metric: Novelty detection when reaction fails.

        Not Implemented since no gt_class info for novel samples

        The method computes novelty detection performance for only on samples with incorrect k-class predictions

        Args:
            p_novel: detection predictions (Dimension: [img X novel])
                Nx1 vector with each element corresponding to probability of novelty
            gt_novel: ground truth detections (Dimension: [img X detection])
                Nx1 vector with each element 0 (not novel) or 1 (novel)
            p_class: detection predictions (Dimension: [img X prob that sample is novel, prob of 88 known classes])
                Nx(K+1) matrix with each row corresponding to K+1 class probabilities for each sample
            gt_class: ground truth classes (Dimension:  [img X classification])
                Nx1 vector with ground-truth class for each sample
            mode: if 'full_test' computes on all test samples, if 'post_novelty' computes from the first GT
                novel sample.  If 'pre_novelty', than everything before novelty introduced.

        Returns:
            Dictionary containing TP, FP, TN, FN, top1, top3 accuracy over the test.
        """
        class_prob = p_class.iloc[:, range(1, p_class.shape[1])].to_numpy()
        gt_class_idx = gt_class.to_numpy()
        return m_ndp_failed_reaction(p_novel, gt_novel, class_prob, gt_class_idx)

    def m_accuracy_on_novel(
        self, p_class: DataFrame, gt_class: DataFrame, gt_novel: DataFrame
    ) -> Dict:
        """
        Additional Metric: Novelty robustness.

        The method computes top-K accuracy for only the novel samples

        Args:
            p_class: detection predictions (Dimension: [img X  prob that sample is novel, prob of known classes])
                Nx(K+1) matrix with each row corresponding to K+1 class probabilities for each sample
            gt_class: ground truth classes (Dimension: [img X classification])
                Nx1 vector with ground-truth class for each sample
            gt_novel: ground truth detections (Dimension: N X [img, classification])
                Nx1 binary vector corresponding to the ground truth novel{1}/seen{0} labels

        Returns:
            Accuracy on novely samples
        """

        class_prob = p_class.iloc[:, range(1, p_class.shape[1])].to_numpy()
        gt_class_idx = gt_class.to_numpy()

        return m_accuracy_on_novel(class_prob, gt_class_idx, gt_novel)

    def m_is_cdt_and_is_early(self, gt_idx: int, ta2_idx: int, test_len: int) -> Dict:
        """
        Is change detection and is change detection early (m_is_cdt_and_is_early) function.

        Args:
            gt_idx: Index when novelty is introduced
            ta2_idx: Index when change is detected
            test_len: Length of test

        Returns
            Dictionary containing boolean showing if change was was detected and if it was detected early
        """
        is_cdt = (ta2_idx >= gt_idx) & (ta2_idx < test_len)
        is_early = ta2_idx < gt_idx
        return {"Is CDT": is_cdt, "Is Early": is_early}

    def m_nrp(self, ta2_acc: Dict, baseline_acc: Dict) -> Dict:
        """
        m_nrp function.

        Args:
            ta2_acc: Accuracy scores for the agent
            baseline_acc: Accuracy scores for baseline

        Returns:
            Reaction performance for the agent
        """
        nrp = {}
        nrp["M_nrp_post_top3"] = 100 * (ta2_acc["post_top3"] / baseline_acc["pre_top3"])
        nrp["M_nrp_post_top1"] = 100 * (ta2_acc["post_top1"] / baseline_acc["pre_top1"])
        return nrp


def convert_df(old_filepath: str, new_filepath: str) -> None:
    """
    Convert from the old df to the new df.

    Args:
        old_filepath:  the filepath the the old *_single_df.csv file
        new_filepath:  the filepath the the old *_single_df.csv file

    Return:
        None
    """
    df = pd.read_csv(old_filepath)
    df["id"] = df.current_path
    df["detection"] = df.cls_novelty.cummax()

    df["classification"] = df.class_id

    # Retain the class num for novel classes but as negative numbers
    mask = df.classification == -1
    df.loc[mask, "classification"] = df.loc[mask, "current_path"].map(
        lambda x: -int(x.split("/")[-2])
    )

    df[["id", "detection", "classification"]].to_csv(new_filepath, index=False)



"""Sail On client feedback package."""

from sail_on_client.feedback.image_classification_feedback import (
    ImageClassificationFeedback,
)
from sail_on_client.feedback.document_transcription_feedback import (
    DocumentTranscriptionFeedback,
)
from sail_on_client.feedback.activity_recognition_feedback import (
    ActivityRecognitionFeedback,
)
from typing import Dict, Union


feedback_type = Union[
    ActivityRecognitionFeedback,
    DocumentTranscriptionFeedback,
    ImageClassificationFeedback,
]


def create_feedback_instance(domain: str, feedback_config: Dict) -> feedback_type:
    """
    Create an instance of feedback object.

    Args:
        domain: Name of the domain
        feedback_config: arguments used to instantiate the metric object
    """
    if domain == "image_classification":
        return ImageClassificationFeedback(**feedback_config)
    elif domain == "activity_recognition":
        return ActivityRecognitionFeedback(**feedback_config)
    elif domain == "transcripts":
        return DocumentTranscriptionFeedback(**feedback_config)
    else:
        raise ValueError(f'Domain: "{domain}" is not a real domain.')



"""Add additional search paths to hydra's default searchpath."""

from hydra.core.config_search_path import ConfigSearchPath
from hydra.plugins.search_path_plugin import SearchPathPlugin


class AdditionalSearchPathPlugin(SearchPathPlugin):
    """Plugin to add additional search paths to hydra's default searchpaths."""

    def manipulate_search_path(self, search_path: ConfigSearchPath) -> None:
        """
        Add additional search paths for configurations for hydra.

        Args:
            search_path: Search path used by hydra

        Returns:
            None
        """
        search_path.append(provider="sail-on-client", path="pkg://hydra_plugins.test")



"""Implementation of T&E Harness for PAR Server."""

import requests
import json
import io
import os
import traceback
import logging

from sail_on_client.harness.test_and_evaluation_harness import TestAndEvaluationHarness
from typing import Any, Dict, Union, List
from requests import Response
from sail_on_client.errors import ApiError, RoundError
from tenacity import (
    retry,
    stop_after_attempt,
    wait_fixed,
    before_sleep_log,
)
from json import JSONDecodeError

log = logging.getLogger(__name__)


class ParHarness(TestAndEvaluationHarness):
    """Harness for PAR server."""

    def __init__(self, url: str, save_directory: str) -> None:
        """
        Initialize a client connection object.

        Args:
            url: URL for the server
            save_directory: A directory to save files

        Returns:
            None
        """
        TestAndEvaluationHarness.__init__(self)
        self.url = url
        self.save_directory = save_directory

    def get_config(self) -> Dict:
        """JSON Compliant representation of the object."""
        return {"url": self.url, "save_directory": self.save_directory}

    def _check_response(self, response: Response) -> None:
        """
        Parse errors that present in the server response.

        Args:
            response: The response object obtained from the server

        Returns:
            None
        """
        if response.status_code != 200:
            try:
                response_json = response.json()
                # Find the appropriate error class based on error code.
                for subclass in ApiError.error_classes():
                    if subclass.error_code == response.status_code:
                        raise subclass(
                            response_json["reason"],
                            response_json["message"],
                            response_json["stack_trace"],
                        )
            except JSONDecodeError:
                log.exception(f"Server Error: {traceback.format_exc()}")
                exit(1)

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_fixed(2),
        reraise=True,
        before_sleep=before_sleep_log(log, logging.INFO),
    )
    def test_ids_request(
        self,
        protocol: str,
        domain: str,
        detector_seed: str,
        test_assumptions: str = "{}",
    ) -> str:
        """
        Request Test Identifiers as part of a series of individual tests.

        Args:
            protocol : string indicating which protocol is being evaluated
            domain : problem domain for the tests
            detector_seed : A seed provided by the novelty detector
            test_assumptions : Assumptions used by the detector

        Returns:
            Filename of file containing test ids
        """
        payload = {
            "protocol": protocol,
            "domain": domain,
            "detector_seed": detector_seed,
        }

        with open(test_assumptions, "r") as f:
            contents = f.read()

        response = requests.get(
            f"{self.url}/test/ids",
            files={  # type: ignore
                "test_requirements": io.StringIO(json.dumps(payload)),
                "test_assumptions": io.StringIO(contents),
            },
        )

        self._check_response(response)

        filename = os.path.abspath(
            os.path.join(
                self.save_directory, f"{protocol}.{domain}.{detector_seed}.csv"
            )
        )
        with open(filename, "w") as f:
            f.write(response.content.decode("utf-8"))

        return filename

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_fixed(2),
        reraise=True,
        before_sleep=before_sleep_log(log, logging.INFO),
    )
    def session_request(
        self,
        test_ids: list,
        protocol: str,
        domain: str,
        novelty_detector_version: str,
        hints: list,
        detection_threshold: float,
    ) -> str:
        """
        Create a new session to evaluate the detector using an empirical protocol.

        Args:
            test_ids: List of tests being evaluated in this session
            protocol: String indicating which protocol is being evaluated
            domain: String indicating which domain is being evaluated
            novelty_detector_version: The novelty detector being evaluated
            hints: Hints used for the session
            detection_threshold: Detection threshold for the session

        Returns:
            A session identifier provided by the server
        """
        payload = {
            "protocol": protocol,
            "novelty_detector_version": novelty_detector_version,
            "domain": domain,
            "hints": hints,
            "detection_threshold": detection_threshold,
        }

        ids = "\n".join(test_ids) + "\n"

        response = requests.post(
            f"{self.url}/session",
            files={"test_ids": ids, "configuration": io.StringIO(json.dumps(payload))},  # type: ignore
        )

        self._check_response(response)
        return response.json()["session_id"]

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_fixed(2),
        reraise=True,
        before_sleep=before_sleep_log(log, logging.INFO),
    )
    def resume_session(self, session_id: str) -> List[str]:
        """
        Get finished test from an existing session.

        Args:
            session id : Session id that was started but not terminated

        Returns:
            List of tests finished in the session
        """
        params: Dict[str, str] = {"session_id": session_id}
        response = requests.get(f"{self.url}/session/latest", params=params)
        self._check_response(response)
        return response.json()["finished_tests"]

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_fixed(2),
        reraise=True,
        before_sleep=before_sleep_log(log, logging.INFO),
    )
    def dataset_request(self, test_id: str, round_id: int, session_id: str) -> str:
        """
        Request data for evaluation.

        Args:
            test_id: The test being evaluated at this moment.
            round_id: The sequential number of the round being evaluated
            session_id: The identifier provided by the server for a single experiment

        Returns:
            Filename of a file containing a list of image files (including full path for each)
        """
        params: Dict[str, Union[str, int]] = {
            "session_id": session_id,
            "test_id": test_id,
            "round_id": round_id,
        }
        response = requests.get(
            f"{self.url}/session/dataset",
            params=params,
        )
        if response.status_code == 204:
            raise RoundError("End of Dataset", "The entire dataset has been requested")
        self._check_response(response)

        filename = os.path.abspath(
            os.path.join(self.save_directory, f"{session_id}.{test_id}.{round_id}.csv")
        )
        with open(filename, "wb") as f:
            f.write(response.content)
        return filename

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_fixed(2),
        reraise=True,
        before_sleep=before_sleep_log(log, logging.INFO),
    )
    def get_feedback_request(
        self,
        feedback_ids: list,
        feedback_type: str,
        test_id: str,
        round_id: int,
        session_id: str,
    ) -> str:
        """
        Get Labels from the server based provided one or more example ids.

        Args:
            feedback_ids: List of media ids for which feedback is required
            feedback_type: Protocols constants with the values: label, detection, characterization
            test_id: The id of the test currently being evaluated
            round_id: The sequential number of the round being evaluated
            session_id: The id provided by a server denoting a session

        Returns:
            Path to a file containing containing requested feedback
        """
        params: Dict[str, Union[str, int]] = {
            "feedback_ids": "|".join(feedback_ids),
            "session_id": session_id,
            "test_id": test_id,
            "feedback_type": feedback_type,
        }
        response = requests.get(
            f"{self.url}/session/feedback",
            params=params,
        )
        self._check_response(response)
        filename = os.path.abspath(
            os.path.join(
                self.save_directory,
                f"{session_id}.{test_id}.{round_id}_{feedback_type}.csv",
            )
        )
        with open(filename, "wb") as f:
            f.write(response.content)

        return filename

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_fixed(2),
        reraise=True,
        before_sleep=before_sleep_log(log, logging.INFO),
    )
    def post_results(
        self, result_files: Dict[str, str], test_id: str, round_id: int, session_id: str
    ) -> None:
        """
        Post client detector predictions for the dataset.

        Args:
            result_files: A dictionary of results with protocol constant as key and file path as value
            test_id: The id of the test currently being evaluated
            round_id: The sequential number of the round being evaluated
            session_id: The id provided by a server denoting a session

        Returns:
            None
        """
        payload = {
            "session_id": session_id,
            "test_id": test_id,
            "round_id": round_id,
            "result_types": "|".join(result_files.keys()),
        }

        files = {"test_identification": io.StringIO(json.dumps(payload))}

        if len(result_files.keys()) == 0:
            raise Exception("Must provide at least one result file")

        for r_type in result_files:
            with open(result_files[r_type], "r") as f:
                contents = f.read()
                files[f"{r_type}_file"] = io.StringIO(contents)

        response = requests.post(f"{self.url}/session/results", files=files)  # type: ignore

        self._check_response(response)

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_fixed(2),
        reraise=True,
        before_sleep=before_sleep_log(log, logging.INFO),
    )
    def evaluate_round_wise(
        self,
        test_id: str,
        round_id: int,
        session_id: str,
    ) -> Dict[str, Any]:
        """
        Get results for round(s).

        Args:
            test_id: The id of the test currently being evaluated
            round_id: The sequential number of the round being evaluated
            session_id: The id provided by a server denoting a session

        Returns:
            Path to a file with the results
        """
        raise NotImplementedError("Round wise accuracy computation is not supported")

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_fixed(2),
        reraise=True,
        before_sleep=before_sleep_log(log, logging.INFO),
    )
    def evaluate(
        self,
        test_id: str,
        round_id: int,
        session_id: str,
        baseline_session_id: str = None,
    ) -> Dict:
        """
        Get results for test(s).

        Args:
            test_id: The id of the test currently being evaluated
            round_id: The sequential number of the round being evaluated
            session_id: The id provided by a server denoting a session

        Returns:
            Path to a file with the results
        """
        params: Dict[str, Union[str, int]] = {
            "session_id": session_id,
            "test_id": test_id,
            "round_id": round_id,
        }
        response = requests.get(
            f"{self.url}/session/evaluations",
            params=params,
        )

        self._check_response(response)
        return json.loads(response.text)

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_fixed(2),
        reraise=True,
        before_sleep=before_sleep_log(log, logging.INFO),
    )
    def get_test_metadata(self, session_id: str, test_id: str) -> Dict[str, Any]:
        """
        Retrieve the metadata json for the specified test.

        Args:
            session_id: The id of the session currently being evaluated
            test_id: The id of the test currently being evaluated

        Returns:
            A dictionary containing metadata
        """
        response = requests.get(
            f"{self.url}/test/metadata",
            params={"test_id": test_id, "session_id": session_id},
        )

        self._check_response(response)
        return response.json()

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_fixed(2),
        reraise=True,
        before_sleep=before_sleep_log(log, logging.INFO),
    )
    def complete_test(self, session_id: str, test_id: str) -> None:
        """
        Mark test as completed.

        Args:
            session_id: The id of the session currently being evaluated
            test_id: The id of the test currently being evaluated

        Returns:
            None
        """
        requests.delete(
            f"{self.url}/test",
            params={"test_id": test_id, "session_id": session_id},
        )

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_fixed(2),
        reraise=True,
        before_sleep=before_sleep_log(log, logging.INFO),
    )
    def terminate_session(self, session_id: str) -> None:
        """
        Terminate the session after the evaluation for the protocol is complete.

        Args:
            session_id: The id provided by a server denoting a session

        Returns: None
        """
        response = requests.delete(
            f"{self.url}/session", params={"session_id": session_id}
        )

        self._check_response(response)



"""Test for CONDDA."""

import logging
from itertools import count
from typing import Dict, List

from sail_on_client.protocol.condda_dataclasses import (
    AlgorithmAttributes,
    InitializeParams,
)
from sail_on_client.protocol.visual_test import VisualTest
from sail_on_client.protocol.condda_round import CONDDARound
from sail_on_client.harness.test_and_evaluation_harness import (
    TestAndEvaluationHarnessType,
)
from sail_on_client.utils.utils import safe_remove
from sail_on_client.errors import RoundError


log = logging.getLogger(__name__)


class CONDDATest(VisualTest):
    """Class representing CONDDA Test."""

    def __init__(
        self,
        algorithm_attributes: AlgorithmAttributes,
        data_root: str,
        domain: str,
        feature_dir: str,
        harness: TestAndEvaluationHarnessType,
        save_dir: str,
        session_id: str,
        skip_stages: List[str],
        use_consolidated_features: bool,
        use_saved_features: bool,
    ) -> None:
        """
        Construct test for CONDDA.

        Args:
            algorithm_attributes: An instance of algorithm_attributes
            data_root: Root directory for the dataset
            domain: Name of the domain for the test
            feature_dir: Directory where features are stored
            harness: An Instance of harness used for T&E
            save_dir: The directory where features are saved
            session_id: Session identifier for the test
            skip_stages: List of stages that would be skipped
            use_consolidated_features: Flag for using consolidated features
            use_saved_features: Flag for using saved features

        Returns:
            None
        """
        super().__init__(
            algorithm_attributes,
            data_root,
            domain,
            feature_dir,
            harness,
            save_dir,
            session_id,
            skip_stages,
            use_consolidated_features,
            use_saved_features,
        )

    def __call__(self, test_id: str) -> None:
        """
        Core logic for running test in CONDDA.

        Args:
            test_id: An identifier for the test

        Returns:
            Score for the test
        """
        metadata = self.harness.get_test_metadata(self.session_id, test_id)
        redlight_instance = metadata.get("red_light", "")
        # Initialize algorithm
        algorithm_instance = self.algorithm_attributes.instance
        algorithm_parameters = self.algorithm_attributes.parameters
        algorithm_init_params = InitializeParams(
            algorithm_parameters, self.session_id, test_id
        )
        algorithm_instance.execute(algorithm_init_params.get_toolset(), "Initialize")

        # Restore features
        features_dict, logit_dict = self._restore_features(test_id)

        # Initialize Round
        round_instance = CONDDARound(
            algorithm_instance,
            self.data_root,
            features_dict,
            self.harness,
            logit_dict,
            redlight_instance,
            self.session_id,
            self.skip_stages,
            test_id,
        )
        aggregated_features_dict: Dict = {}
        aggregated_logit_dict: Dict = {}
        # Run algorithm for multiple rounds
        for round_id in count(0):
            log.info(f"Start round: {round_id}")
            # see if there is another round available
            try:
                dataset = self.harness.dataset_request(
                    test_id, round_id, self.session_id
                )
            except RoundError:
                # no more rounds available, this test is done.
                break
            round_instance(dataset, round_id)
            (
                aggregated_features_dict,
                aggregated_logit_dict,
            ) = self._aggregate_features_across_round(
                round_instance, aggregated_features_dict, aggregated_logit_dict
            )
            # cleanup the dataset file for the round
            safe_remove(dataset)
            log.info(f"Round complete: {round_id}")
        self.harness.complete_test(self.session_id, test_id)
        self._save_features(test_id, aggregated_features_dict, aggregated_logit_dict)



"""Abstract class for feedback for sail-on."""

import pandas as pd
from sail_on_client.harness.par_harness import ParHarness
from sail_on_client.harness.local_harness import LocalHarness

from typing import Union, Dict


class Feedback:
    """Base class for Feedback."""

    def __init__(
        self,
        first_budget: int,
        income_per_batch: int,
        maximum_budget: int,
        interface: Union[LocalHarness, ParHarness],
        session_id: str,
        test_id: str,
        feedback_type: str,
    ) -> None:
        """
        Initialize.

        Args:
            first_budget: Initial budget
            income_per_batch: Additional labels added after every batch
            maximum_budget: Max labels that can be requested
            interface: An instance of evaluation interface
            session_id: Session identifier
            test_id: Test identifier
            feedback_type: Type of feedback that can be requested

        Returns:
            None
        """
        self.budget = first_budget
        self.income_per_batch = income_per_batch
        self.maximum_budget = maximum_budget
        self.current_round = -1
        self.interface = interface
        self.session_id = session_id
        self.test_id = test_id
        self.feedback_type = feedback_type

    def get_labeled_feedback(
        self, round_id: int, images_id_list: list, image_names: list
    ) -> Union[pd.DataFrame, None]:
        """
        Get labeled feedback for the round.

        Args:
            round_id: Round identifier
            image_id_list: List if indices for images
            image_names: List of image names for the round

        Return:
            A dictionary with the accuracy value or None if
            feedback is requested for an older round
        """
        if round_id > self.current_round:
            self.deposit_income()
            self.current_round = round_id
            if len(images_id_list) <= self.budget:
                self.budget = self.budget - len(images_id_list)
                image_ids = [image_names[int(idx)] for idx in images_id_list]
                feedback_file = self.interface.get_feedback_request(
                    image_ids,
                    self.feedback_type,
                    self.test_id,
                    round_id,
                    self.session_id,
                )

                df = pd.read_csv(
                    feedback_file, delimiter=",", header=None, names=["id", "labels"]
                )
            else:
                raise ValueError("the function should be added")
        else:
            df = None
        return df

    def get_score_feedback(
        self, round_id: int, images_id_list: list, image_names: list
    ) -> Union[Dict, None]:
        """
        Get accuracy value for the round. Note: this is not budgeted.

        Args:
            round_id: Round identifier
            image_id_list: List if indices for images
            image_names: List of image names for the round

        Return:
            A dictionary with the accuracy value or None if
            feedback is requested for an older round
        """
        if round_id > self.current_round:
            self.deposit_income()
            self.current_round = round_id
            image_ids = [image_names[int(idx)] for idx in images_id_list]
            feedback_file = self.interface.get_feedback_request(
                image_ids,
                self.feedback_type,
                self.test_id,
                round_id,
                self.session_id,
            )
            df = pd.read_csv(feedback_file, delimiter=",", header=None)
            return df
        else:
            return None

    def get_feedback(
        self, round_id: int, images_id_list: list, image_names: list
    ) -> Union[pd.DataFrame, Dict, None]:
        """
        Get feedback for the round.

        Args:
            round_id: Round identifier
            image_id_list: List if indices for images
            image_names: List of image names for the round

        Return:
            Either a dataframe or dictionary with score if the request is valid
            for the current round.
        """
        if self.feedback_type == "classification":
            feedback_fn = self.get_labeled_feedback
        elif self.feedback_type == "score":
            feedback_fn = self.get_score_feedback
        else:
            raise ValueError("Unsupported feedback type {self.feedback_type} specified")
        return feedback_fn(round_id, images_id_list, image_names)

    def deposit_income(self) -> None:
        """Get income for a round."""
        self.budget = min(self.maximum_budget, (self.budget + self.income_per_batch))

    def get_budget(self) -> int:
        """Get current budget."""
        return self.budget



"""Helper functions for metrics."""

import numpy as np
import pandas as pd
from typing import List


def check_novel_validity(p_novel: np.ndarray, gt_novel: np.ndarray) -> None:
    """
    Check the validity of the inputs for per-sample novelty detection.

    Args:
        p_novel: NX1 vector with each element corresponding to probability of novelty
        gt_novel: NX1 vector with each element 0 (not novel) or 1 (novel)

    Returns:
        None
    """
    if p_novel.shape[0] != gt_novel.shape[0]:
        raise Exception(
            "Number of predicted samples not equal to number of groundtruth samples!"
        )
    if p_novel.ndim != 1:
        raise Exception(
            "Predicted probabilities must be a vector but is an array of dimension {}!".format(
                p_novel.ndim
            )
        )
    if np.any(p_novel < 0) or np.any(p_novel > 1):
        raise Exception("Predicted novel probabilities should be between 0 and 1!")
    if np.any(np.logical_and(gt_novel != 0, gt_novel != 1)):
        raise Exception(
            "Groundtruth array should only consist of 0s (non-novel) or 1s(novel)!"
        )
    return


def check_class_validity(p_class: np.ndarray, gt_class: np.ndarray) -> None:
    """
    Check the validity of the inputs for image classification.

    Inputs:
    p_class: Nx(K+1) matrix with each row corresponding to K+1 class probabilities for each sample
    gt_class: Nx1 vector with ground-truth class for each sample
    """
    if p_class.shape[0] != gt_class.shape[0]:
        raise Exception(
            "Number of predicted samples not equal to number of groundtruth samples!"
        )
    if np.any(p_class < 0) or np.any(p_class > 1):
        raise Exception("Predicted class probabilities should be between 0 and 1!")
    if p_class.ndim != 2:
        raise Exception(
            "Predicted probabilities must be a 2D matrix but is an array of dimension {}!".format(
                p_class.ndim
            )
        )
    if np.max(gt_class) >= p_class.shape[1] or np.min(gt_class < 0):
        raise Exception(
            "Groundtruth class labels must lie in the range [0-{}]!".format(
                p_class.shape[1]
            )
        )
    return


def topk_accuracy(
    p_class: np.ndarray, gt_class: np.ndarray, k: int, txt: str = ""
) -> float:
    """
    Compute top-K accuracy.

    Args:
        p_class: Nx(K+1) matrix with each row corresponding to K+1 class probabilities for each sample
        gt_class: Nx1 computevector with ground-truth class for each sample
        k: 'k' used in top-K accuracy
        txt: Text associated with accuracy

    Returns:
        top-K accuracy
    """
    check_class_validity(p_class, gt_class)
    p_class = np.argsort(-p_class)[:, :k]
    gt_class = gt_class[:, np.newaxis]
    check_zero: np.ndarray = p_class - gt_class
    correct = np.sum(np.any(check_zero == 0, axis=1).astype(int))
    return round(float(correct) / p_class.shape[0], 5)


def top3_accuracy(p_class: np.ndarray, gt_class: np.ndarray, txt: str = "") -> float:
    """
    Compute top-3 accuracy. (see topk_accuracy() for details).

    Args:
        p_class: Nx(K+1) matrix with each row corresponding to K+1 class probabilities for each sample
        gt_class: Nx1 computevector with ground-truth class for each sample
        txt: Text associated with accuracy

    Returns:
        top-3 accuracy
    """
    return topk_accuracy(p_class, gt_class, k=3, txt=txt)


def top1_accuracy(p_class: np.ndarray, gt_class: np.ndarray, txt: str = "") -> float:
    """
    Compute top-1 accuracy. (see topk_accuracy() for details).

    Args:
        p_class: Nx(K+1) matrix with each row corresponding to K+1 class probabilities for each sample
        gt_class: Nx1 computevector with ground-truth class for each sample
        txt: Text associated with accuracy

    Returns:
        top-1 accuracy
    """
    return topk_accuracy(p_class, gt_class, k=1, txt=txt)


# compute information for the robustness measures
def get_rolling_stats(
    p_class: np.ndarray, gt_class: np.ndarray, k: int = 1, window_size: int = 50
) -> List:
    """
    Compute rolling statistics which are used for robustness measures.

    Args:
        p_class: Nx(K+1) matrix with each row corresponding to K+1 class probabilities for each sample
        gt_class: Nx1 compute vector with ground-truth class for each sample
        k: 'k' used for selecting top k values
        window_size: Window size for running stats

    Returns:
        List with mean and standard deviation
    """
    p_cls_topk = np.argsort(-p_class)[:, :k]
    gt_cls_indx = gt_class[:, np.newaxis]
    check_zero_topk = p_cls_topk - gt_cls_indx
    topk_correct = 1 * (np.any(check_zero_topk == 0, axis=1))
    acc_mean = pd.Series(topk_correct).rolling(window=window_size).mean().mean()
    acc_std = pd.Series(topk_correct).rolling(window=window_size).mean().std()
    return [acc_mean, acc_std]


def get_first_detect_novelty(p_novel: np.ndarray, thresh: float) -> int:
    """
    Find the first index where novelty is detected.

    Args:
        p_novel: NX1 vector with each element corresponding to probability of novelty
        thresh: Score threshold for detecting when a sample is novel

    Returns:
        Index where an agent reports that a sample is novel
    """
    if np.sum(p_novel >= thresh) < 1:
        first_detect_novelty = len(p_novel) + 1
    else:
        first_detect_novelty = np.where(p_novel >= thresh)[0][0] + 1
    return first_detect_novelty



"""Set of functions used by file provider."""
from sail_on_client.errors.errors import ServerError, ProtocolError, RoundError
from sail_on_client.harness.constants import ProtocolConstants

import csv
import datetime
import numpy as np
import nltk
import os
import json
from typing import List, Optional, Dict, Any
from sklearn.metrics.cluster import normalized_mutual_info_score
import re
import traceback


def read_gt_csv_file(file_location: str) -> List:
    """
    Read the ground truth csv file.

    Args:
        file_location: Path to the gt csv file

    Returns:
        List of entries in csv file
    """
    with open(file_location, "r") as f:
        csv_reader = csv.reader(f, delimiter=",", quotechar="|")
        return list(csv_reader)[1:]


def read_meta_data(file_location: str) -> Dict:
    """
    Read the metadata file.

    Args:
        file_location: Path to the gt csv file

    Returns:
        Content of the metadata file
    """
    with open(file_location, "r") as md:
        return json.load(md)


def get_encoding(domain: str) -> str:
    """
    Get encoding based on the domain.

    Args:
        Name of the domain

    Returns:
        Encoding string for the domain
    """
    if domain == "nlt":
        return ProtocolConstants.NLT_ENCODING
    elif domain == "activity_recognition":
        return ProtocolConstants.VAR_ENCODING
    elif domain == "transcripts":
        return ProtocolConstants.WTR_ENCODING
    else:
        return "utf-8"


def get_session_info(
    folder: str, session_id: str, in_process_only: bool = True
) -> Dict[str, Any]:
    """
    Retrieve session info.

    Args:
        folder: Folder where the session file is saved
        session_id: Session id for which the info is required
        in_process_only: Flag to get information while the session is active

    Returns:
        Session information as a dict
    """
    path = os.path.join(folder, f"{str(session_id)}.json")
    if os.path.exists(path):
        with open(path, "r") as session_file:
            info = json.load(session_file)
            terminated = "termination" in info
            if terminated and in_process_only:
                raise ProtocolError(
                    "SessionEnded",
                    """The session being requested has already been terminated.
                       Please either create a new session or request a different ID""",
                )
            else:
                return info
    return {}


def get_session_test_info(folder: str, session_id: str, test_id: str) -> Dict[str, Any]:
    """
    Retrieve session info for a test.

    Args:
        folder: Folder where the session file is saved
        session_id: Session id for which the info is required
        test_id: Test id for which the info is required

    Returns:
        Session information associated with test as a dict
    """
    path = os.path.join(folder, f"{str(session_id)}.{test_id}.json")
    if os.path.exists(path):
        with open(path, "r") as session_file:
            info = json.load(session_file)
            if "completion" in info:
                raise ProtocolError(
                    "TestCompleted",
                    "The test being requested has already been completed for this session",
                )
            else:
                return info
    return {}


def write_session_log_file(structure: Dict, filepath: str) -> None:
    """
    Write session information in a log file.

    Args:
        structure: Dict with information about the session
        filepath: Path to the file where log file is saved

    Returns:
        None
    """
    with open(filepath, "w") as session_file:
        json.dump(structure, session_file, indent=2)


def log_session(
    folder: str,
    session_id: str,
    activity: str,
    test_id: Optional[str] = None,
    round_id: Optional[int] = None,
    content: Optional[Dict[str, Any]] = None,
    content_loc: Optional[str] = "round",
    return_structure: Optional[bool] = False,
) -> Optional[Dict]:
    """
    Create a log files of all session activity.

    Args:
        folder: Folder where the session file is saved
        session_id: Session id associated with the log
        test_id: Test id associated with the log
        round_id: Round id associated with the log
        content: Content of the log
        content_loc: Location where the content should be added
        return_structure: Flag to return the updated log

    Returns
        Updated log if `return_structure` is set to True
    """
    structure = get_session_info(folder, session_id)
    write_session_file = True
    if test_id is None:
        structure[activity] = {"time": [str(datetime.datetime.now())]}
        if content is not None:
            structure[activity].update(content)
    else:
        test_structure = get_session_test_info(folder, session_id, test_id)
        if activity not in test_structure:
            test_structure[activity] = {"time": [str(datetime.datetime.now())]}
        if content_loc == "activity":
            if content is not None:
                test_structure[activity].update(content)
        if round_id is not None:
            round_id_str = str(round_id)
            rounds = test_structure[activity].get("rounds", {})
            if round_id_str not in rounds:
                rounds[round_id_str] = {"time": [str(datetime.datetime.now())]}
            else:
                rounds[round_id_str]["time"].append(str(datetime.datetime.now()))
            if content_loc == "round":
                if content is not None:
                    rounds[round_id_str].update(content)
            test_structure[activity]["rounds"] = rounds
            test_structure[activity]["last round"] = round_id_str

        if not return_structure:
            write_session_log_file(
                test_structure,
                os.path.join(folder, f"{str(session_id)}.{str(test_id)}.json"),
            )

        if activity == "completion":
            session_tests = structure.get("tests", {"completed_tests": []})
            session_tests["completed_tests"].append(test_id)
            structure["tests"] = session_tests
        else:
            write_session_file = False

    if write_session_file:
        write_session_log_file(
            structure, os.path.join(folder, f"{str(session_id)}.json")
        )

    if return_structure:
        return test_structure
    return None


def read_feedback_file(
    csv_reader: "csv.reader",  # type: ignore
    feedback_ids: Optional[List[str]],
    metadata: Dict[str, Any],
    check_constrained: bool = True,
) -> Dict:
    """
    Get feedback from feedback file for the specified ids in the last submitted round.

    Args:
        csv_reader: An instance of csv reader
        feedback_ids: Element ids for which feedback is requested
        metadata: Dictionary with metadata
        check_constrained: Flag to check constraints associated with feedback

    Returns:
        Dictionary containing feedback with feedback_ids as keys
    """
    feedback_constrained = metadata.get("feedback_constrained", True)

    lines: List = list(csv_reader)

    try:
        if not check_constrained or not feedback_constrained:
            start = 0
            end = len(lines)
        else:
            # under the constrained case, we always look at the last round
            start = len(lines) - int(metadata["round_size"])
            end = start + int(metadata["round_size"])
    except KeyError:
        raise RoundError(
            "no_defined_rounds",
            "round_size not defined in metadata.",
            "".join(traceback.format_stack()),
        )

    if feedback_ids:
        return {
            x[0]: list(x[1:])
            for x in [[n.strip(" \"'") for n in y] for y in lines][start:end]
            if x[0] in feedback_ids
        }
    else:
        return {
            x[0]: list(x[1:])
            for x in [[n.strip(" \"'") for n in y] for y in lines][start:end]
        }


def get_classification_feedback(
    gt_file: str,
    result_files: List[str],
    feedback_ids: List[str],
    metadata: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Calculate the proper feedback for classification type feedback.

    Args:
        gt_file: Path to ground truth file
        result_files: List of paths with results
        feedback_ids: Element ids for which feedback is requested
        metadata: Dictionary with metadata

    Returns:
        Dictionary containing feedback with feedback_ids as keys
    """
    if feedback_ids is None or len(feedback_ids) == 0:
        # if feedback ids not provided, limit to those in the last round
        with open(result_files[0], "r") as rf:
            result_reader = csv.reader(rf, delimiter=",")
            results = read_feedback_file(
                result_reader, None, metadata, check_constrained=True
            )
            feedback_max_ids = min(
                metadata.get("feedback_max_ids", len(results)), len(results)
            )
            feedback_ids = list(results.keys())[: int(feedback_max_ids)]

    ground_truth = read_feedback_file(
        read_gt_csv_file(gt_file),
        feedback_ids,
        metadata,
        check_constrained=feedback_ids is None or len(feedback_ids) == 0,
    )

    return {
        x: min(int(ground_truth[x][metadata["columns"][0]]), metadata["known_classes"])
        for x in ground_truth.keys()
    }


def get_kinetics_labels_var_feedback(
    gt_file: str,
    result_files: List[str],
    feedback_ids: List[str],
    metadata: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Get feedback for video activity recognition.

    Args:
        gt_file: Path to ground truth file
        result_files: List of paths with results
        feedback_ids: Element ids for which feedback is requested
        metadata: Dictionary with metadata

    Returns:
        Dictionary containing feedback with feedback_ids as keys
    """
    ground_truth = read_feedback_file(
        read_gt_csv_file(gt_file),
        feedback_ids,
        metadata,
        check_constrained=feedback_ids is None or len(feedback_ids) == 0,
    )

    return {
        x: ground_truth[x][metadata["columns"][0] : metadata["columns"][1]]
        for x in ground_truth.keys()
    }


def get_single_gt_feedback(
    gt_file: str,
    result_files: List[str],
    feedback_ids: List[str],
    metadata: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Get feedback that specifies if a sample if correct or incorrect.

    Args:
        gt_file: Path to ground truth file
        result_files: List of paths with results
        feedback_ids: Element ids for which feedback is requested
        metadata: Dictionary with metadata

    Returns:
        Dictionary containing feedback with feedback_ids as keys
    """
    with open(result_files[0], "r") as rf:
        result_reader = csv.reader(rf, delimiter=",")
        results = read_feedback_file(
            result_reader, feedback_ids, metadata, check_constrained=True
        )

    # if feedback ids not provided, limit ids grabbed from ground truth to the last submitted round
    gt_feedback_ids = []
    if feedback_ids is None or len(feedback_ids) == 0:
        gt_feedback_ids = list(results.keys())
    else:
        gt_feedback_ids = feedback_ids
    ground_truth = read_feedback_file(
        read_gt_csv_file(gt_file),
        gt_feedback_ids,
        metadata,
        check_constrained=gt_feedback_ids is None or len(gt_feedback_ids) == 0,
    )

    return_ids = []
    # For specific feedback types, and when no feedback ids are specified,
    # will only return feedback on instances marked incorrectly
    if metadata.get("return_incorrect", None) and not feedback_ids:
        for sample_id in results.keys():
            r = -1
            if metadata["return_incorrect"] == ProtocolConstants.CLASSIFICATION:
                r = int(np.argmax([float(i) for i in results[sample_id]], axis=0))
            elif metadata["return_incorrect"] == ProtocolConstants.DETECTION:
                r = int(results[sample_id][1])
            else:
                raise ProtocolError(
                    "FeedbackConfigError",
                    "The api based feedback config is misconfigured. Please check API",
                )
            g = int(ground_truth[sample_id][metadata["columns"][0]])
            if r != g:
                return_ids.append(sample_id)
    else:
        return_ids = list(ground_truth.keys())

    return {x: ground_truth[x][metadata["columns"][0]] for x in return_ids}


def get_classificaton_score_feedback(
    gt_file: str,
    result_files: List[str],
    feedback_ids: List[str],
    metadata: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Calculate feedback on the accuracy of classification.

    Args:
        gt_file: Path to ground truth file
        result_files: List of paths with results
        feedback_ids: Element ids for which feedback is requested
        metadata: Dictionary with metadata

    Returns:
        Dictionary containing feedback with feedback_ids as keys
    """
    ground_truth = read_feedback_file(
        read_gt_csv_file(gt_file), None, metadata, check_constrained=False
    )
    with open(result_files[0], "r") as rf:
        result_reader = csv.reader(rf, delimiter=",")
        results = read_feedback_file(
            result_reader, None, metadata, check_constrained=False
        )

    # Go through results and count number correct
    num_correct = 0
    for idx in results.keys():
        r = int(np.argmax([float(i) for i in results[idx]], axis=0))
        g = int(ground_truth[idx][metadata["columns"][0]])
        if r == g:
            num_correct += 1

    accuracy = float(num_correct) / float(len(results.keys()))
    return {"accuracy": accuracy}


def get_characterization_feedback(
    gt_file: str,
    result_files: List[str],
    feedback_ids: List[str],
    metadata: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Calculate the proper feedback for characterization type feedback.

    Args:
        gt_file: Path to ground truth file
        result_files: List of paths with results
        feedback_ids: Element ids for which feedback is requested
        metadata: Dictionary with metadata

    Returns:
        Dictionary containing feedback with feedback_ids as keys
    """
    # Not implemented
    raise NameError("Characterization Feedback is not supported.")
    known_classes = int(metadata["known_classes"]) + 1

    with open(result_files[0], "r") as rf:
        result_reader = csv.reader(rf, delimiter=",")
        results = read_feedback_file(result_reader, feedback_ids, metadata)
    ground_truth = read_feedback_file(
        read_gt_csv_file(gt_file), feedback_ids, metadata, check_constrained=False
    )

    # If ground truth is not novel, returns 1 is prediction is correct,
    # otherwise returns 1 if prediction is not a known class
    return {
        x: 0
        if (
            sum(ground_truth[x][1:known_classes])
            > (ground_truth[x][0] + sum(ground_truth[x][known_classes:]))
            and ground_truth[x].index(max(ground_truth[x]))
            != results[x].index(max(results[x]))
        )
        or (
            sum(ground_truth[x][1:known_classes])
            < (ground_truth[x][0] + sum(ground_truth[x][known_classes:]))
            and (results[x].index(max(results[x])) in range(1, known_classes))
        )
        else 1
        for x in ground_truth.keys()
    }


def _ensure_space(input_str: str) -> str:
    return " ".join(
        [
            x.strip()
            for x in re.split(
                r"(\W+)",
                input_str.replace(";", "")
                .replace('"', "")
                .replace("|", "")
                .replace("  ", " "),
            )
        ]
    )


def get_levenshtein_feedback(
    gt_file: str,
    result_files: List[str],
    feedback_ids: List[str],
    metadata: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Calculate the proper feedback for levenshtein type feedback.

    Args:
        gt_file: Path to ground truth file
        result_files: List of paths with results
        feedback_ids: Element ids for which feedback is requested
        metadata: Dictionary with metadata

    Returns:
        Dictionary containing feedback with feedback_ids as keys
    """
    ground_truth = read_feedback_file(
        read_gt_csv_file(gt_file), feedback_ids, metadata, check_constrained=False
    )
    with open(result_files[0], "r") as rf:
        result_reader = csv.reader(rf, delimiter=",")
        results = read_feedback_file(result_reader, feedback_ids, metadata)

    return {
        x: [
            nltk.edit_distance(
                _ensure_space(ground_truth[x][metadata["columns"][i]]),
                _ensure_space(results[x][0]),
            )
            for i, _ in enumerate(metadata["columns"])
        ]
        for x in results.keys()
    }


def get_cluster_feedback(
    gt_file: str,
    result_files: List[str],
    feedback_ids: List[str],
    metadata: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Calculate the proper feedback for cluster type feedback.

    Args:
        gt_file: Path to ground truth file
        result_files: List of paths with results
        feedback_ids: Element ids for which feedback is requested
        metadata: Dictionary with metadata

    Returns:
        Dictionary containing feedback with feedback_ids as keys
    """
    ground_truth = read_feedback_file(
        read_gt_csv_file(gt_file), feedback_ids, metadata, check_constrained=False
    )
    with open(result_files[0], "r") as rf:
        result_reader = csv.reader(rf, delimiter=",")
        results = read_feedback_file(result_reader, feedback_ids, metadata)

    if feedback_ids is None:
        feedback_ids = ground_truth.keys()

    # clear ground truth of all but relevant columns
    for key in ground_truth.keys():
        ground_truth[key] = [ground_truth[key][i] for i in metadata["columns"]]

    gt_list = []
    r_list = []
    try:
        for key in sorted(feedback_ids):
            gt_list.append([float(x) for x in ground_truth[key]])
            r_list.append([float(x) for x in results[key]])
    except Exception:
        raise ServerError(
            "MissingIds",
            """Some requested Ids are missing from either ground truth or results
            file for the current round""",
        )

    gt_np = np.array(gt_list).reshape(len(gt_list))
    r_np = np.argmax(np.array(r_list), axis=1)

    return_dict = {"nmi": normalized_mutual_info_score(gt_np, r_np)}

    for i in np.unique(r_np):
        places = np.where(r_np == i)[0]
        return_dict[str(i)] = (
            max(np.unique(gt_np[places], return_counts=True)[1]) / places.shape[0]
        )

    return return_dict


def psuedo_label_feedback(
    gt_file: str,
    feedback_ids: List[str],
    feedback_type: str,
    metadata: Dict[str, Any],
    folder: str,
    session_id: str,
) -> Dict[str, Any]:
    """
    Get psuedo label feedback for requested ids.

    Args:
        gt_file: Path to ground truth file
        feedback_ids: Element ids for which feedback is requested
        feedback_type: Type of feedback
        metadata: Dictionary with metadata
        folder: Folder where the session file is saved
        session_id: Session id for which the info is required

    Returns:
        Dictionary containing feedback with feedback_ids as keys
    """
    ground_truth = read_feedback_file(read_gt_csv_file(gt_file), feedback_ids, metadata)

    structure = get_session_info(folder, session_id)

    if "psuedo_labels" in structure:
        if feedback_type in structure["psuedo_labels"]:
            labels = structure["psuedo_labels"][feedback_type]
        else:
            structure["psuedo_labels"][feedback_type] = []
            labels = []
    else:
        structure["psuedo_labels"] = {feedback_type: []}
        labels = []

    return_dict = {}
    for x in ground_truth.keys():
        col = int(ground_truth[x][metadata["columns"][0]])
        if col not in labels:
            labels.append(col)
        return_dict[x] = labels.index(col)

    structure["psuedo_labels"][feedback_type] = labels
    write_session_log_file(structure, os.path.join(folder, f"{str(session_id)}.json"))

    return return_dict



"""Utility function for sail-on-client."""

import os
import logging

from typing import Dict, List

log = logging.getLogger(__name__)


def safe_remove(file_path: str) -> None:
    """
    Remove a file after checking that it exists.

    Args:
        file_path (str): File path that should be removed

    Return:
        None
    """
    if os.path.exists(file_path):
        os.remove(file_path)


def safe_remove_results(results: dict) -> None:
    """
    Remove results present in a dict.

    Args:
        file_path (str): File path that should be removed

    Return:
        None
    """
    for result_files in results.values():
        safe_remove(result_files)


def merge_dictionaries(base_dict: Dict, other_dict: Dict, exclude_keys: List) -> Dict:
    """
    Shallow merge for two dictionaries with certain keys would be skipped.

    Args:
        base_dict: Dictionary with defaults
        other_dict: Dictionary with arguments that are added or overriden
        exclude_keys: Keys in other dict that shouldn't be used in the merge

    Returns:
        Dictionary with parameters obtained by merging two dictionaries
    """
    merged_dict = base_dict.copy()
    included_keys = set(other_dict.keys()).difference(set(exclude_keys))
    for key in included_keys:
        merged_dict[key] = other_dict[key]
    return merged_dict



"""Common dataclasses for visual protocols."""

from dataclasses import dataclass
import logging
from typing import Dict


log = logging.getLogger(__name__)


@dataclass
class FeatureExtractionParams:
    """Class for storing feature extraction parameters associated with an algorithm."""

    dataset: str
    data_root: str
    round_id: int

    def get_toolset(self) -> Dict:
        """
        Convert the data present in the class into a dictionary.

        Returns
            A dictionary with data associated with the class
        """
        return {
            "dataset": self.dataset,
            "dataset_root": self.data_root,
            "round_id": self.round_id,
        }


@dataclass
class WorldChangeDetectionParams:
    """Class for storing parameters associated world change in an algorithm."""

    features_dict: Dict
    logit_dict: Dict
    round_id: int
    redlight_image: str

    def get_toolset(self) -> Dict:
        """
        Convert the data present in the class into a dictionary.

        Returns
            A dictionary with data associated with the class
        """
        return {
            "features_dict": self.features_dict,
            "logit_dict": self.logit_dict,
            "round_id": self.round_id,
            "redlight_image": self.redlight_image,
        }



"""Abstract harness for T&E."""

from smqtk_core import Configurable, Pluggable
from abc import abstractmethod
from typing import List, Dict, Any, TypeVar

TestAndEvaluationHarnessType = TypeVar(
    "TestAndEvaluationHarnessType", bound="TestAndEvaluationHarness"
)


class TestAndEvaluationHarness(Configurable, Pluggable):
    """Abstract interface for test and evaluation harness."""

    @classmethod
    def is_usable(cls) -> bool:
        """Determine if this class with be detected by SMQTK's plugin."""
        return True

    def get_config(self) -> Dict:
        """Return a default configuration dictionary."""
        return {}

    @abstractmethod
    def test_ids_request(
        self,
        protocol: str,
        domain: str,
        detector_seed: str,
        test_assumptions: str = "{}",
    ) -> str:
        """
        Request Test Identifiers as part of a series of individual tests.

        Args:
            protocol : string indicating which protocol is being evaluated
            domain : problem domain for the tests
            detector_seed : A seed provided by the novelty detector
            test_assumptions : Assumptions used by the detector

        Returns:
            Filename of file containing test ids
        """
        pass

    @abstractmethod
    def session_request(
        self,
        test_ids: list,
        protocol: str,
        domain: str,
        novelty_detector_version: str,
        hints: list,
        detection_threshold: float,
    ) -> str:
        """
        Create a new session to evaluate the detector using an empirical protocol.

        Args:
            test_ids: List of tests being evaluated in this session
            protocol: String indicating which protocol is being evaluated
            domain: String indicating which domain is being evaluated
            novelty_detector_version: The novelty detector being evaluated
            hints: Hints used for the session
            detection_threshold: Detection threshold for the session

        Returns:
            A session identifier provided by the server
        """
        pass

    @abstractmethod
    def resume_session(self, session_id: str) -> List[str]:
        """
        Get finished test from an existing session.

        Args:
            session id : Session id that was started but not terminated

        Returns:
            List of tests finished in the session
        """
        pass

    @abstractmethod
    def dataset_request(self, test_id: str, round_id: int, session_id: str) -> str:
        """
        Request data for evaluation.

        Args:
            test_id: The test being evaluated at this moment.
            round_id: The sequential number of the round being evaluated
            session_id: The identifier provided by the server for a single experiment

        Returns:
            Filename of a file containing a list of image files (including full path for each)
        """
        pass

    @abstractmethod
    def get_feedback_request(
        self,
        feedback_ids: list,
        feedback_type: str,
        test_id: str,
        round_id: int,
        session_id: str,
    ) -> str:
        """
        Get Labels from the server based provided one or more example ids.

        Args:
            feedback_ids: List of media ids for which feedback is required
            feedback_type: Protocols constants with the values: label, detection, characterization
            test_id: The id of the test currently being evaluated
            round_id: The sequential number of the round being evaluated
            session_id: The id provided by a server denoting a session

        Returns:
            Path to a file containing containing requested feedback
        """
        pass

    @abstractmethod
    def post_results(
        self, result_files: Dict[str, str], test_id: str, round_id: int, session_id: str
    ) -> None:
        """
        Post client detector predictions for the dataset.

        Args:
            result_files: A dictionary of results with protocol constant as key and file path as value
            test_id: The id of the test currently being evaluated
            round_id: The sequential number of the round being evaluated
            session_id: The id provided by a server denoting a session

        Returns:
            None
        """
        pass

    @abstractmethod
    def evaluate_round_wise(
        self,
        test_id: str,
        round_id: int,
        session_id: str,
    ) -> Dict[str, Any]:
        """
        Get results for round(s).

        Args:
            test_id: The id of the test currently being evaluated
            round_id: The sequential number of the round being evaluated
            session_id: The id provided by a server denoting a session

        Returns:
            Path to a file with the results
        """
        pass

    @abstractmethod
    def evaluate(
        self,
        test_id: str,
        round_id: int,
        session_id: str,
        baseline_session_id: str = None,
    ) -> Dict:
        """
        Get results for test(s).

        Args:
            test_id: The id of the test currently being evaluated
            round_id: The sequential number of the round being evaluated
            session_id: The id provided by a server denoting a session

        Returns:
            Path to a file with the results
        """
        pass

    @abstractmethod
    def get_test_metadata(self, session_id: str, test_id: str) -> Dict[str, Any]:
        """
        Retrieve the metadata json for the specified test.

        Args:
            session_id: The id of the session currently being evaluated
            test_id: The id of the test currently being evaluated

        Returns:
            A dictionary containing metadata
        """
        pass

    @abstractmethod
    def complete_test(self, session_id: str, test_id: str) -> None:
        """
        Mark test as completed.

        Args:
            session_id: The id of the session currently being evaluated
            test_id: The id of the test currently being evaluated

        Returns:
            None
        """
        pass

    @abstractmethod
    def terminate_session(self, session_id: str) -> None:
        """
        Terminate the session after the evaluation for the protocol is complete.

        Args:
            session_id: The id provided by a server denoting a session

        Returns: None
        """
        pass



"""Sail On client evaluate package."""

from sail_on_client.evaluate.image_classification import ImageClassificationMetrics
from sail_on_client.evaluate.document_transcription import DocumentTranscriptionMetrics
from sail_on_client.evaluate.activity_recognition import ActivityRecognitionMetrics
from typing import Dict, Union


metric_type = Union[
    ActivityRecognitionMetrics, DocumentTranscriptionMetrics, ImageClassificationMetrics
]


def create_metric_instance(protocol: str, domain: str, gt_config: Dict) -> metric_type:
    """
    Create an instance of metric object.

    Args:
        protocol: Name of the protocol
        domain: Name of the domain
        gt_config: Additional arguments used by the metric object
    """
    if domain == "image_classification":
        return ImageClassificationMetrics(protocol, **gt_config)
    elif domain == "activity_recognition":
        return ActivityRecognitionMetrics(protocol, **gt_config)
    elif domain == "transcripts":
        return DocumentTranscriptionMetrics(protocol, **gt_config)
    else:
        raise ValueError(f'Domain: "{domain}" is not a real domain.')
